#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <sys/stat.h>

#define TRUE       1
#define FALSE      0
#define maxfields 20

int 	maxwidth;

struct	fldinfo	
{
	char	recdfield[13];     /* field name of database record */
	int 	pos;               /* index to start of field */
	int 	lng;               /* length of field */
} ;

void explain(void)
{
  fprintf(stderr, "SUMMARY   generates summary information from PC-FILE+ databases.   It produces\n");
  fprintf(stderr, "  subtotal counts on specified fields,  telling how many records in a row have\n");
  fprintf(stderr, "  the same value for a specified field.  If more than one field is given, then\n");
  fprintf(stderr, "  sub-sub-totals (if that's a real phrase) are given, etc.   The database must\n");
  fprintf(stderr, "  have been sorted by PC-FILE+ with the same primary sort field (any secondary\n");
  fprintf(stderr, "  fields too) in the same sequence they are given on the SUMMARY call line for\n");
  fprintf(stderr, "  the sub-totals to be meaningful.\n\n");
  fprintf(stderr, "Usage: SUMMARY database [fldname [fldname2 ...]] [>output]\n");
  fprintf(stderr, "  where:  database  is the name of the PC-FILE+ database excluding\n");
  fprintf(stderr, "                    the file extension\n");
  fprintf(stderr, "          fldname   is the primary sort field for subtotals\n");
  fprintf(stderr, "          fldname2, etc are secondary sort field names\n");
  fprintf(stderr, "    if no field name is given, SUMMARY shows the database structure\n");
  fprintf(stderr, "          output    is the name of the file into which the summary\n");
  fprintf(stderr, "                    statistics are to be written.   If redirection\n");
  fprintf(stderr, "                    is not specified, output will go to the screen\n\n");
}

int hdread(char *dbasename, struct fldinfo fieldefs[], int *recdlng)
{
	char	hdrfilnm[30];
	char	line[256];
	int 	fieldlength, fieldct, endfile;
	FILE	*hdrf;
	char	fieldname[13];

  strcpy(hdrfilnm, dbasename);
  strcat(hdrfilnm, ".HDR");   /* create the database header file name */
  if ((hdrf = fopen(hdrfilnm, "r")) < 0 )
  {
    perror("Error on open of database definition file: ");
    return(0);
  }
  endfile = FALSE;
  for (fieldct = 0, *recdlng = 0; !endfile; )
  {             /* put one line's worth into line */
    endfile = ( fscanf(hdrf, "%[^\n]\n", &line) == EOF || line[0] == '|' );
    if (!endfile)                     /* the screen layout begins with '|' */
    {
      sscanf(line, "%12s", &fieldname);            /* strip out field name */
      endfile = ( fscanf(hdrf, "%[^\n]\n", &line) == EOF );
      if (!endfile)                    /* really shouldn't ever get an EOF */
      {
        sscanf(line, "%d", &fieldlength);            /* get field's length */
        strcpy(fieldefs[fieldct].recdfield, fieldname);
        fieldefs[fieldct].pos = *recdlng;  /* current recd length is start */
        fieldefs[fieldct++].lng = fieldlength;
        *recdlng += fieldlength;                 /* maintain record length */
      }
    }
  } /* end of for loop */
  (*recdlng)++;     /* There's a Carriage Return at end of each dbase recd */
  fclose(hdrf);                        /* must include it in record length */
  return(fieldct);        /* pass back number of fields in database record */
} /* end of hdread */

void bar(int curfld, int fldnm[], struct fldinfo defs[], char *last[],
         int excess)
{   /* This prints a bar of just the right length before the subtotal line */
	int 	i, j;

  for (i=0; i<excess; i++) printf("--");
  for (i=curfld; i>=0; i--)
  {
    printf("----");
    for (j=0; j<strlen(defs[fldnm[i]].recdfield); j++) putchar('-');
    for (j=0; j<strlen(last[i]); j++) putchar('-');
  }
  for (i=0; i<maxwidth; i++) putchar('-');
  printf("-----------\n");
}

void sumfield(int curfld, int fldnm[], struct fldinfo defs[],
              char *last[], int reccnt)
{
	int 	i;

  printf("%*u records have ", maxwidth, reccnt);
  for (i=curfld; i>=0; i--)
  {
    printf("%s=%s", defs[fldnm[i]].recdfield, last[i]);
    if (i>0) printf(" & ");
  }
  printf("\n");
}

main(int argc, char *argv[])
{
	struct	fldinfo	fieldefs[maxfields];
	char	dbasename[30], inxfile[30], dtafile[30];
	int 	inxf, dtaf;
	int 	fieldct, fieldnum[maxfields], fieldndx, numfields, i, j;
	int 	recdlng, brkhit, brkpt[maxfields], recdcount[maxfields];
	unsigned int totrecs, recdptr;
	char	fieldname[13];
	char	*last[maxfields], *inxrecd, *dbrec;
	long	seekadr;
	int 	fn, iorslt;
	char	filelngstr[33];
	long	maxrecds;

  if (argc<2)            /* If called with no parameters, tell how to call */
  {
    explain();
    return;
  }
  strcpy(dbasename, argv[1]);   /* database name is first argument on call */
  strupr(dbasename);
  fieldct = hdread(dbasename, fieldefs, &recdlng);
  if (fieldct == 0)
  {
    fprintf(stderr, "Invalid database: %s, aborting\n\n", dbasename);
    explain();
    return;
  }
  strcpy(inxfile, dbasename);
  strcat(inxfile, ".INX");              /* create database index file name */
  inxrecd = malloc(fieldct*2);   /* get space for the index record */
  strcpy(dtafile, dbasename);
  strcat(dtafile, ".DTA");               /* create database data file name */
  if (argc<3)  /* if no field names are given, describe database structure */
  {
    fprintf(stderr, "\nFollowing is the database record structure for %s\n\n",
            dbasename);
    for (i=0; i<fieldct; i++)
      fprintf(stderr, "Field # %d = %12s is %3d bytes long, starting at byte # %d\n",
              i+1, fieldefs[i].recdfield, fieldefs[i].lng, fieldefs[i].pos);
    if ( (inxf = open(inxfile, O_RDONLY | O_BINARY)) < 0 )
    {
      perror("\nError on open of database index file: ");
      return;
    }
    if ( filelength(inxf) < (long) (fieldct + 1)*4 )
      fprintf(stderr, "\nIndex file is empty or improper format\n\n");
    else if ( lseek(inxf, -(2*fieldct)-2, SEEK_END) < 0L )
      fprintf(stderr, "nError on seek in index file\n\n");
    else if ( read(inxf, inxrecd, fieldct*2) < 1 )
      fprintf(stderr, "\nError on read of index file\n\n");
    else if ( read(inxf, (char*) &recdptr, sizeof(int)) < 1 )
      fprintf(stderr, "\nError on read of index file\n\n");
    else if ( inxrecd[0] != '\\' )  /* Last record begins with '\' /*
      fprintf(stderr, "nIndex file is improper format\n\n");
    else if ( inxrecd[2] == '\xFF' )  /* Field number of primary key */
      fprintf(stderr, "\nDatabase is unsorted\n");
    else
      fprintf(stderr, "\nDatabase is sorted by %s\n",
                      fieldefs[inxrecd[2]].recdfield);
    if ( (dtaf = open(dtafile, O_RDONLY | O_BINARY)) < 0 )
    {
      perror("Error on open of database data file: ");
      return;
    }
    fprintf(stderr, "Database has %ld records (some may be deleted)\n",
                    filelength(inxf)/(long)(fieldct*2+2)-1);
    close(inxf);
    close(dtaf);
    return;
  }
  dbrec = malloc(recdlng);    /* get space for the database record */
  numfields = 0;
  for (fieldndx = 2; fieldndx < argc; fieldndx++)
  {
    strncpy(fieldname, argv[fieldndx], 12);     /* read in the field names */
    for (i=0; (i<fieldct) &&
         (strnicmp(fieldefs[i].recdfield, fieldname, strlen(fieldname)) != 0) ;
          i++) ;               /* search for match in database field names */
    if (i >= fieldct)
      fprintf(stderr, "%s does not match a field's name in the database\n", fieldname);
    else fieldnum[numfields++] = i;
  }
  fprintf(stderr, "Number of matched fields = %d\n", numfields);
  if (numfields == 0)
  {
    fprintf(stderr, "You have not matched any fieldnames of the database--aborting.\n");
    return;
  }
  for (i=0; i<numfields; i++)                /* tell about selected fields */
  {                                 /* and initialize necessary data areas */
    fn = fieldnum[i];
    fprintf(stderr, "Field #%d is %s\n", i+1,
            fieldefs[fn].recdfield);
    last[i] = (char*) malloc(fieldefs[fn].lng+1); /* get space for last rec */
    last[i][0] = 0; /* a field at a time */  /* start value as null string */
    last[i][fieldefs[fn].lng] = 0;      /* guarantee an end of string code */
    recdcount[i] = 0;    /* each field has a record counter for sub-totals */
  }
  if ( (inxf = open(inxfile, O_RDONLY | O_BINARY)) < 0)
  {
    perror("Error on open of database index file: ");
    return;
  }
  strcpy(dtafile, dbasename);
  strcat(dtafile, ".DTA");               /* create database data file name */
  if ( (dtaf = open(dtafile, O_RDONLY | O_BINARY)) < 0)
  {
    perror("Error on open of database data file: ");
    return;
  }
  maxrecds = filelength(inxf) / (long)(fieldct*2+2) - 1;
  ltoa(maxrecds, filelngstr, 10);
  maxwidth = strlen(filelngstr);
  printf("Summary statistics for database %s\n\n", dbasename);
  inxrecd[0] = 0;
  totrecs = 0;
  while ( (read(inxf, inxrecd, fieldct*2) > 0) && (inxrecd[0] != '\\'))
  { /* loop until end of file hit */
    if (inxrecd[0] != '/')               /* deleted records begin with '/' */
    {
      if (read(inxf, (char*) &recdptr, sizeof(int)) < 1) break;
      brkhit = FALSE;  /* flag to tell if need a sub-total */
           /* locate data record in database file */
      if ( lseek(dtaf, (long)(recdptr-1)*recdlng, SEEK_SET) < 0L )
      {
        perror("Error on seek of database index file: ");
        break;
      }
      if ( (iorslt = read(dtaf, dbrec, recdlng)) == 0)
      {
        perror("Error reading database data file: ");
        break;
      }
      if (iorslt < 0) break;
      for (i=0; i<numfields; i++)
      { /* loop through checking if hit a breaking point for any field */
        fn = fieldnum[i];
        brkhit |= (totrecs>0) &&  /* if a break point hit earlier, break */
          (strncmp(dbrec+fieldefs[fn].pos, last[i], fieldefs[fn].lng) != 0);
        brkpt[i] = brkhit;  /*  if mismatches prior records field, break */
      }
      for (i=numfields-1; i>=0; i--)
      { /* working downwards, check for breakpoints */
        if ((recdcount[i] > 0) && brkpt[i])
        { /* We need a sub-total, have something to count */
          if (i<numfields-1) bar(i, fieldnum, fieldefs, last, numfields-i-1);
          for (j=0; j<numfields-i-1; j++) printf("  ");          /* indent */
          sumfield(i, fieldnum, fieldefs, last, recdcount[i]);
          if (i<numfields-1) printf("\n");
          recdcount[i] = 0;
        }
        fn = fieldnum[i];            /* copy current record fields to last */
        strncpy(last[i], dbrec+fieldefs[fn].pos, fieldefs[fn].lng);
        recdcount[i]++;                 /* we have a live record--count it */
      }
      totrecs++;                       /* data base record counter */
    }
  } /* end of reading data base records */
  if ( totrecs > 0 )          /* clean up leftover sub-totals */
    for (i=numfields-1; i>=0; i--)
    {
      if (i<numfields-1) bar(i, fieldnum, fieldefs, last, numfields-i-1);
      for (j=0; j<numfields-i-1; j++) printf("  ");
      sumfield(i, fieldnum, fieldefs, last, recdcount[i]);
      if (i<numfields-1) printf("\n");
    }
  for (i=0; i<maxwidth; i++) putchar('=');
  printf("=============================================\n\n");
  printf("There are a total of %d records in the database\n", totrecs);
  close(inxf);
  close(dtaf);
}
